# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime, date, timedelta, timezone
from decimal import Decimal
from typing import Any, Mapping

from .. import LOG
from ..model.core import Party
from ..model.types_store import PackageStore
from ..util.prim_types import frozendict
from .daml_lf_1 import *


class Evaluator:

    def __init__(
            self,
            store: 'PackageStore',
            value_bindings: 'Mapping[str, Expr]',
            type_bindings: 'Mapping[str, Type]'):
        self.store = store
        self.value_bindings = frozendict(value_bindings)
        self.type_bindings = frozendict(type_bindings)

    def child_evaluator(
            self,
            value_bindings: 'Optional[Mapping[str, Expr]]' = None,
            type_bindings: 'Optional[Mapping[str, Type]]' = None) -> 'Evaluator':
        if value_bindings:
            new_value_bindings = dict()
            new_value_bindings.update(self.value_bindings)
            vars_to_delete = []
            for k, v in value_bindings.items():
                if not isinstance(k, str):
                    raise TypeError()
                if not isinstance(v, (Expr, Evaluator.Constant)):
                    raise TypeError('Only Expr and Evaluator.Constant are allowed here')
                if isinstance(v, Expr):
                    if v.var is not None and v.var == k:
                        vars_to_delete.append(v.var)
                if k in self.value_bindings:
                    LOG.warning('Variable %r is shadowing another variable with the same name in an outer scope! (original value: %r, new value: %r)', k, self.value_bindings[k], v)

            for k in vars_to_delete:
                del value_bindings[k]

            new_value_bindings.update(value_bindings)
        else:
            new_value_bindings = self.value_bindings

        if type_bindings:
            new_type_bindings = dict()
            new_type_bindings.update(self.type_bindings)
            new_type_bindings.update(type_bindings)
        else:
            new_type_bindings = self.type_bindings

        return Evaluator(self.store, new_value_bindings, new_type_bindings)

    class Constant:
        def __init__(self, value):
            if isinstance(value, Expr):
                raise TypeError('The Constant wrapper is only intended to be used with non-Expr')
            self.value = value

        def __repr__(self):
            return f'Constant({self.value!r})'

    def eval_Expr(self, expr: 'Expr') -> Any:
        if isinstance(expr, Evaluator.Constant):
            return expr.value
        result = expr.Sum_match(
            self.eval_var,
            self.eval_val,
            self.eval_builtin,
            self.eval_prim_con,
            self.eval_prim_lit,
            self.eval_rec_con,
            self.eval_rec_proj,
            self.eval_variant_con,
            self.eval_tuple_con,
            self.eval_tuple_proj,
            self.eval_app,
            self.eval_ty_app,
            self.eval_abs,
            self.eval_ty_abs,
            self.eval_case,
            self.eval_let,
            self.eval_nil,
            self.eval_cons,
            self.eval_update,
            self.eval_scenario,
            self.eval_rec_upd,
            self.eval_tuple_upd,
            self.eval_none,
            self.eval_some)
        if isinstance(result, Expr):
            LOG.warning('An Expr eval generated another Expr, which is probably incorrect.')
            LOG.warning('  Initial Expr: %r', expr)
            LOG.warning('  Result Expr: %r', result)
        return result

    def eval_var(self, var: str) -> Any:
        expr = self.value_bindings.get(var)
        if expr is None:
            LOG.warning('var %r could not be resolved to a value in the current scope!', var)
            return Expr(var=var)

        return self.eval_Expr(expr)

    def eval_val(self, val: 'ValName') -> 'Any':
        if val.full_name_unambiguous == 'DA.Internal.Prelude:concat':
            return PartialFunction(lambda a: PartialFunction(Evaluator.concat))
        value = self.store.resolve_value_reference(val)
        if value is None:
            LOG.warning('data type %s is undefined!', val)
            return Expr(val=val)

        return self.eval_Expr(value)

    @staticmethod
    def concat(xxs):
        result = []
        for xs in xxs:
            result.extend(xs)
        return result

    @staticmethod
    def eval_builtin(builtin: 'BuiltinFunction') -> 'Any':
        return BUILTINS.definition(builtin)

    @staticmethod
    def eval_prim_con(prim_con: 'PrimCon') -> 'Any':
        if PrimCon.CON_TRUE == prim_con:
            return True
        elif PrimCon.CON_FALSE == prim_con:
            return False
        elif PrimCon.CON_UNIT == prim_con:
            return {}
        else:
            raise ValueError(f'unknown PrimCon value: {prim_con}')

    @staticmethod
    def eval_prim_lit(prim_lit: 'PrimLit') -> 'Any':
        if prim_lit.int64 is not None:
            return prim_lit.int64
        elif prim_lit.decimal is not None:
            return Decimal(prim_lit.decimal)
        elif prim_lit.text is not None:
            return prim_lit.text
        elif prim_lit.timestamp is not None:
            return datetime.fromtimestamp(prim_lit.timestamp / 1e9, timezone.utc)
        elif prim_lit.party is not None:
            return Party(prim_lit.party)
        elif prim_lit.date is not None:
            return date(1970, 1, 1) + timedelta(days=prim_lit.date)
        else:
            raise ValueError(f'unknown PrimLit value: {prim_lit}')

    def eval_rec_con(self, rec_con: 'Expr.RecCon') -> 'Any':
        return frozendict({fwt.field: self.eval_Expr(fwt.expr) for fwt in rec_con.fields})

    def eval_rec_proj(self, rec_proj: 'Expr.RecProj') -> 'Any':
        record_expr = self.eval_Expr(rec_proj.record)
        return record_expr[rec_proj.field]

    def eval_variant_con(self, variant_con: 'Expr.VariantCon') -> 'Any':
        ctor = variant_con.variant_con
        value = self.eval_Expr(variant_con.variant_arg)
        return frozendict({ctor: value})

    def eval_tuple_con(self, tuple_con: 'Expr.TupleCon') -> 'Any':
        return frozendict({fwt.field: self.eval_Expr(fwt.expr) for fwt in tuple_con.fields})

    def eval_tuple_proj(self, tuple_proj: 'Expr.TupleProj') -> 'Any':
        tuple_expr = self.eval_Expr(tuple_proj.tuple)
        return tuple_expr[tuple_proj.field]

    def eval_app(self, app: 'Expr.App') -> 'Any':
        fun = self.eval_Expr(app.fun)
        if hasattr(fun, 'kind') and fun.kind != 'abs':
            raise ValueError('oh what the hell')

        args = [self.eval_Expr(arg) for arg in app.args]
        return fun.apply(*args)

    def eval_ty_app(self, ty_app: 'Expr.TyApp') -> 'Any':
        fun = self.eval_Expr(ty_app.expr)
        if hasattr(fun, 'kind') and fun.kind != 'ty_abs':
            raise ValueError('oh what the hell')
        # TODO: No actual Type visiting happens here; is this necessary for us?
        args = [arg for arg in ty_app.types]
        return fun.apply(*args)

    def eval_abs(self, abs_: 'Expr.Abs') -> 'Any':
        func_parameters = [p.var for p in abs_.param]

        pf = PartialFunction(
            PartialFunctionBody(self, func_parameters, abs_.body),
            arg_count=len(func_parameters))
        pf.kind = 'abs'
        return pf

    def eval_ty_abs(self, ty_abs: 'Expr.TyAbs') -> 'Any':
        ty_func_parameters = [t.var for t in ty_abs.param]

        # the body will return a Callable, but only when all variables are bound.
        # Delay evaluation of the expression until all variables' values are gathered
        def implementation(*args):
            evaluator = self.child_evaluator(type_bindings=dict(
                zip(ty_func_parameters, args)))
            return evaluator.eval_Expr(ty_abs.body)

        pf = PartialFunction(implementation, arg_count=len(ty_func_parameters))
        pf.kind = 'ty_abs'
        return pf

    def eval_case(self, case: 'Case') -> 'Any':
        scrut = self.eval_Expr(case.scrut)
        raise ValueError('not handling case yet')

    def eval_let(self, let: 'Block') -> 'Any':
        current_scope = self
        for binding in let.bindings:
            var = binding.binder.var
            value = current_scope.eval_Expr(binding.bound)
            current_scope = current_scope.child_evaluator(value_bindings={var: Evaluator.Constant(value)})
        return current_scope.eval_Expr(let.body)

    @staticmethod
    def eval_nil(nil: 'Expr.Nil') -> 'Any':
        return ()

    def eval_cons(self, cons: 'Expr.Cons') -> 'Any':
        result = []
        for f in cons.front:
            resolved_val = self.eval_Expr(f)
            try:
                result.append(resolved_val)
            except:
                LOG.exception()
                raise
        result.extend(self.eval_Expr(cons.tail))
        return tuple(result)

    def eval_update(self, update: 'Update') -> 'Any':
        raise ValueError('not handling updates yet')

    def eval_scenario(self, scenario: 'Scenario') -> 'Any':
        raise ValueError('not handling scenarios yet')

    def eval_rec_upd(self, rec_upd: 'Expr.RecUpd') -> 'Any':
        raise ValueError('not handling updates yet')

    def eval_tuple_upd(self, tuple_upd: 'Expr.TupleUpd') -> 'Any':
        raise ValueError('not handling updates yet')

    @staticmethod
    def eval_none(none: '') -> 'Any':
        return None

    def eval_some(self, some: 'Expr.Some') -> 'Any':
        return self.eval_Expr(some.body)


class BuiltinImpl:
    def __init__(self):
        self._functions = {
            BuiltinFunction.APPEND_TEXT: PartialFunction(self.append_text),
            BuiltinFunction.TO_TEXT_INT64: PartialFunction(self.to_text_int64),
            BuiltinFunction.TO_TEXT_PARTY: PartialFunction(self.to_text_party),
            BuiltinFunction.FOLDL: PartialFunction(self.foldl),
            BuiltinFunction.FOLDR: PartialFunction(self.foldr),
        }

    def definition(self, builtin: 'BuiltinFunction') -> 'Callable[[...], Any]':
        return self._functions[builtin]

    def append_text(self, xs, ys) -> str:
        return xs + ys

    def to_text_int64(self, int64) -> str:
        return str(int64)

    def to_text_party(self, party) -> str:
        return party

    def foldl(self, func, acc, xs):
        from functools import reduce
        return reduce(func, acc, xs)

    def foldr(self, func, acc, xs):
        from functools import reduce
        return reduce(lambda x, y: func(y, x), xs[::-1], acc)


class PartialFunction:
    """
    Wrapper around a function that accumulates arguments and invokes the underlying function only
    once all arguments have been collected.
    """
    def __init__(self, fn, *args, arg_count: 'Optional[int]' = None):
        from inspect import signature, Parameter

        if arg_count is None:
            allowed_param_kinds = (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
            # try to compute argument count from the function
            parameters = signature(fn).parameters.values()
            if not all(p.kind in allowed_param_kinds for p in parameters):
                raise ValueError(f'the function {fn!r} cannot be used in a PartialFunction because '
                                 f'it takes an unknown number of parameters; specify arg_count '
                                 f'explicitly if you wish to use the function')
            arg_count = len(parameters)

        if arg_count == 0:
            raise ValueError('Non-zero number of arguments required for a PartialFunction')
        if len(args) >= arg_count:
            raise ValueError('YOU GOT MORE WORK TO DO BUDDY')

        self.fn = fn
        self.args = args
        self.arg_count = arg_count

    def apply(self, *args):
        """
        Apply the specified number of arguments to this function, possibly invoking the underlying
        function and returning its result.

        :param args:
            The arguments to apply to this function.
        :return:
            Either a new :class:`PartialFunction` instance with some values applied, or the result
            of invoking the underlying function with the accumulated values.
        """
        total_args = self.args + args
        total_arg_count = len(total_args)
        if total_arg_count < self.arg_count:
            pf = PartialFunction(self.fn, total_args, arg_count=self.arg_count)
            if hasattr(self, 'kind'):
                pf.kind = self.kind
            return pf
        elif total_arg_count == self.arg_count:
            return self.fn(*total_args)
        else:
            # fully apply this function; THEN expect the function to return a function, and apply
            # remaining arguments to THAT
            remaining_args = total_arg_count - self.arg_count

            result = self.fn(*total_args[:-remaining_args])
            return result(*total_args[remaining_args:])

    def __call__(self, *args):
        return self.apply(*args)

    def __repr__(self):
        return f'PartialFunction({self.fn!r}, args={self.args}, arg_count={self.arg_count})'


class PartialFunctionBody:
    def __init__(self, scope: Evaluator, func_parameters: Sequence[str], func_body_expr: 'Expr'):
        self.scope = scope
        self.func_parameters = tuple(func_parameters)
        self.func_body_expr = func_body_expr

    def __call__(self, *args):
        new_value_bindings = dict(zip(self.func_parameters, [Evaluator.Constant(arg) for arg in args]))
        evaluator = self.scope.child_evaluator(value_bindings=new_value_bindings)
        return evaluator.eval_Expr(self.func_body_expr)

    def __repr__(self):
        return f'PartialFunctionBody(func_parameters={self.func_parameters}, func_body={self.func_body_expr})'


BUILTINS = BuiltinImpl()



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import time_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_testing_dot_time__service__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class TimeServiceStub(object):
  """Optional service, exposed for testing static time scenarios.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetTime = channel.unary_stream(
        '/com.digitalasset.ledger.api.v1.testing.TimeService/GetTime',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_testing_dot_time__service__pb2.GetTimeRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_testing_dot_time__service__pb2.GetTimeResponse.FromString,
        )
    self.SetTime = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.testing.TimeService/SetTime',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_testing_dot_time__service__pb2.SetTimeRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )


class TimeServiceServicer(object):
  """Optional service, exposed for testing static time scenarios.
  """

  def GetTime(self, request, context):
    """Returns a stream of time updates.
    Always returns at least one response, where the first one is the current time.
    Subsequent responses are emitted whenever the ledger server's time is updated.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetTime(self, request, context):
    """Allows clients to change the ledger's clock in an atomic get-and-set operation.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_TimeServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetTime': grpc.unary_stream_rpc_method_handler(
          servicer.GetTime,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_testing_dot_time__service__pb2.GetTimeRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_testing_dot_time__service__pb2.GetTimeResponse.SerializeToString,
      ),
      'SetTime': grpc.unary_unary_rpc_method_handler(
          servicer.SetTime,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_testing_dot_time__service__pb2.SetTimeRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.testing.TimeService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/ledger_configuration_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import trace_context_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/ledger_configuration_service.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B$LedgerConfigurationServiceOuterClass'),
  serialized_pb=_b('\nAcom/digitalasset/ledger/api/v1/ledger_configuration_service.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a\x32\x63om/digitalasset/ledger/api/v1/trace_context.proto\x1a\x1egoogle/protobuf/duration.proto\"x\n\x1dGetLedgerConfigurationRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"s\n\x1eGetLedgerConfigurationResponse\x12Q\n\x14ledger_configuration\x18\x01 \x01(\x0b\x32\x33.com.digitalasset.ledger.api.v1.LedgerConfiguration\"m\n\x13LedgerConfiguration\x12*\n\x07min_ttl\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\x12*\n\x07max_ttl\x18\x02 \x01(\x0b\x32\x19.google.protobuf.Duration2\xb8\x01\n\x1aLedgerConfigurationService\x12\x99\x01\n\x16GetLedgerConfiguration\x12=.com.digitalasset.ledger.api.v1.GetLedgerConfigurationRequest\x1a>.com.digitalasset.ledger.api.v1.GetLedgerConfigurationResponse0\x01\x42\x46\n\x1e\x63om.digitalasset.ledger.api.v1B$LedgerConfigurationServiceOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,google_dot_protobuf_dot_duration__pb2.DESCRIPTOR,])




_GETLEDGERCONFIGURATIONREQUEST = _descriptor.Descriptor(
  name='GetLedgerConfigurationRequest',
  full_name='com.digitalasset.ledger.api.v1.GetLedgerConfigurationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetLedgerConfigurationRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetLedgerConfigurationRequest.trace_context', index=1,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=185,
  serialized_end=305,
)


_GETLEDGERCONFIGURATIONRESPONSE = _descriptor.Descriptor(
  name='GetLedgerConfigurationResponse',
  full_name='com.digitalasset.ledger.api.v1.GetLedgerConfigurationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_configuration', full_name='com.digitalasset.ledger.api.v1.GetLedgerConfigurationResponse.ledger_configuration', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=307,
  serialized_end=422,
)


_LEDGERCONFIGURATION = _descriptor.Descriptor(
  name='LedgerConfiguration',
  full_name='com.digitalasset.ledger.api.v1.LedgerConfiguration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_ttl', full_name='com.digitalasset.ledger.api.v1.LedgerConfiguration.min_ttl', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_ttl', full_name='com.digitalasset.ledger.api.v1.LedgerConfiguration.max_ttl', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=424,
  serialized_end=533,
)

_GETLEDGERCONFIGURATIONREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETLEDGERCONFIGURATIONRESPONSE.fields_by_name['ledger_configuration'].message_type = _LEDGERCONFIGURATION
_LEDGERCONFIGURATION.fields_by_name['min_ttl'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_LEDGERCONFIGURATION.fields_by_name['max_ttl'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
DESCRIPTOR.message_types_by_name['GetLedgerConfigurationRequest'] = _GETLEDGERCONFIGURATIONREQUEST
DESCRIPTOR.message_types_by_name['GetLedgerConfigurationResponse'] = _GETLEDGERCONFIGURATIONRESPONSE
DESCRIPTOR.message_types_by_name['LedgerConfiguration'] = _LEDGERCONFIGURATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetLedgerConfigurationRequest = _reflection.GeneratedProtocolMessageType('GetLedgerConfigurationRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETLEDGERCONFIGURATIONREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.ledger_configuration_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetLedgerConfigurationRequest)
  ))
_sym_db.RegisterMessage(GetLedgerConfigurationRequest)

GetLedgerConfigurationResponse = _reflection.GeneratedProtocolMessageType('GetLedgerConfigurationResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETLEDGERCONFIGURATIONRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.ledger_configuration_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetLedgerConfigurationResponse)
  ))
_sym_db.RegisterMessage(GetLedgerConfigurationResponse)

LedgerConfiguration = _reflection.GeneratedProtocolMessageType('LedgerConfiguration', (_message.Message,), dict(
  DESCRIPTOR = _LEDGERCONFIGURATION,
  __module__ = 'com.digitalasset.ledger.api.v1.ledger_configuration_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.LedgerConfiguration)
  ))
_sym_db.RegisterMessage(LedgerConfiguration)


DESCRIPTOR._options = None

_LEDGERCONFIGURATIONSERVICE = _descriptor.ServiceDescriptor(
  name='LedgerConfigurationService',
  full_name='com.digitalasset.ledger.api.v1.LedgerConfigurationService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=536,
  serialized_end=720,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetLedgerConfiguration',
    full_name='com.digitalasset.ledger.api.v1.LedgerConfigurationService.GetLedgerConfiguration',
    index=0,
    containing_service=None,
    input_type=_GETLEDGERCONFIGURATIONREQUEST,
    output_type=_GETLEDGERCONFIGURATIONRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_LEDGERCONFIGURATIONSERVICE)

DESCRIPTOR.services_by_name['LedgerConfigurationService'] = _LEDGERCONFIGURATIONSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains models used on the read-side of the Ledger API.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Collection, Optional, TypeVar, Union

from .core import ContractId, ContractData, ContractContextualData, Party
from .lookup import template_reverse_globs, validate_template
from .types import Type, TypeReference
from .types_store import PackageStore
from ..util.typing import safe_cast, safe_optional_cast


T = TypeVar('T')


class BaseEvent:
    """
    Superclass of all dazl events.
    """
    def __init__(self, client, party, time, ledger_id, package_store):
        self.client = client
        self.party = safe_cast(str, party) if party is not None else None  # type: Optional[Party]
        self.time = safe_optional_cast(datetime, time)
        self.ledger_id = safe_cast(str, ledger_id)
        self.package_store = safe_cast(PackageStore, package_store)

    def acs_find_active(self, template: Union[TypeReference, str], match=None):
        return self.client.find_active(template, match)

    def acs_find_by_id(self, cid: Union[str, ContractId]) -> Optional[ContractContextualData]:
        return self.client.find_by_id(cid)

    def acs_find_one(self, template: Union[TypeReference, str], match=None):
        return self.client.find_one(template, match=match)

    def acs_find_historical(self, template: Union[TypeReference, str], match=None):
        return self.client.find_historical(template, match)

    def acs_find_nonempty(self, template: Union[TypeReference, str], match=None):
        return self.client.find_nonempty(template, match=match)

    def __repr__(self):
        fields = ', '.join(f"{k}={v!r}" for k, v in self.__dict__.items()
                           if not k.startswith('_') and k != 'client' and k != 'package_store' and
                           k != 'ledger_id')
        return f'{self.__class__.__name__}({fields})'


class InitEvent(BaseEvent):
    """
    Event raised when dazl is initialized, but before it has begun reading from the Active Contract
    Set (ACS).
    """


class OffsetEvent(BaseEvent):
    """
    Event raised when dazl is ready to begin processing new events. At this point, the Active
    Contract Set (ACS) is populated with the current state of the ledger.
    """
    def __init__(self, client, party, time, ledger_id, package_store, offset):
        super().__init__(client, party, time, ledger_id, package_store)
        self.offset = safe_cast(str, offset)


class ReadyEvent(OffsetEvent):
    """
    Event raised when dazl is ready to begin processing new events. At this point, the Active
    Contract Set (ACS) is populated with the current state of the ledger.
    """


class BaseTransactionEvent(OffsetEvent):
    """
    Event raised when dazl encounters a new transaction. This is raised before any corresponding
    :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """
    def __init__(self, client, party, time, ledger_id, package_store, offset, command_id,
                 workflow_id):
        super().__init__(client, party, time, ledger_id, package_store, offset)
        self.command_id = safe_cast(str, command_id)
        self.workflow_id = safe_cast(str, workflow_id)


class TransactionStartEvent(BaseTransactionEvent):
    """
    Event raised when dazl encounters a new transaction. This is raised before any corresponding
    :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """
    def __init__(self, client, party, time, ledger_id, package_store, offset, command_id,
                 workflow_id, contract_events):
        super().__init__(client, party, time, ledger_id, package_store, offset, command_id,
                         workflow_id)
        self.contract_events = contract_events


class TransactionEndEvent(BaseTransactionEvent):
    """
    Event raised when dazl encounters the end of a transaction. This is raised after any
    corresponding :class:`ContractCreateEvent` or :class:`ContractArchiveEvent`.
    """
    def __init__(self, client, party, time, ledger_id, package_store, offset, command_id,
                 workflow_id, contract_events):
        super().__init__(client, party, time, ledger_id, package_store, offset, command_id,
                         workflow_id)
        self.contract_events = contract_events


class ContractEvent(BaseTransactionEvent):
    """
    Event raised when dazl automation detects a new create or an archive. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """
    def __init__(self, client, party, time, ledger_id, package_store, offset, cid, cdata,
                 command_id, workflow_id, event_id, witness_parties):
        super().__init__(client, party, time, ledger_id, package_store, offset, command_id,
                         workflow_id)
        self.cid = safe_cast(ContractId, cid)
        self.cdata = cdata  # type: ContractData
        self.command_id = command_id
        self.workflow_id = workflow_id
        self.event_id = event_id
        self.witness_parties = witness_parties


class ContractCreateEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract create. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """


class ContractExercisedEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract exercised.
    """
    def __init__(self, client, party, time, ledger_id, package_store, offset, cid, cdata,
                 command_id, workflow_id, event_id, witness_parties, contract_creating_event_id,
                 choice, choice_args, acting_parties, consuming, child_event_ids):
        super().__init__(client, party, time, ledger_id, package_store, offset, cid, cdata,
                         command_id, workflow_id, event_id, witness_parties)
        self.contract_creating_event_id = contract_creating_event_id
        self.choice = choice
        self.choice_args = choice_args
        self.acting_parties = acting_parties
        self.consuming = consuming
        self.child_event_ids = child_event_ids


class ContractArchiveEvent(ContractEvent):
    """
    Event raised when dazl automation detects a contract archive. The Active Contract Set
    (ACS) reflects this event, as well as all other events that occurred in the same transaction.
    """


def create_dispatch(
        on_init: Callable[[InitEvent], T],
        on_ready: Callable[[ReadyEvent], T],
        on_offset: Callable[[OffsetEvent], T],
        on_transaction_start: Callable[[TransactionStartEvent], T],
        on_transaction_end: Callable[[TransactionEndEvent], T],
        on_contract_created: Callable[[ContractCreateEvent], T],
        on_contract_exercised: Callable[[ContractExercisedEvent], T],
        on_contract_archived: Callable[[ContractArchiveEvent], T]) -> Callable[[BaseEvent], T]:
    def handle(event: BaseEvent) -> T:
        if isinstance(event, ContractCreateEvent):
            return on_contract_created(event)
        elif isinstance(event, ContractExercisedEvent):
            return on_contract_exercised(event)
        elif isinstance(event, ContractArchiveEvent):
            return on_contract_archived(event)
        elif isinstance(event, TransactionStartEvent):
            return on_transaction_start(event)
        elif isinstance(event, TransactionEndEvent):
            return on_transaction_end(event)
        elif isinstance(event, ReadyEvent):
            return on_ready(event)
        elif isinstance(event, InitEvent):
            return on_init(event)
        elif isinstance(event, OffsetEvent):
            return on_offset(event)
        else:
            raise ValueError(f'unknown subclass of BaseEvent: {event!r}')

    return handle


@dataclass(frozen=True)
class TransactionFilter:
    ledger_id: str
    current_offset: Optional[str]
    destination_offset: Optional[str]
    templates: Optional[Collection[Type]]
    max_blocks: Optional[int]
    party_groups: Optional[Collection[str]]


class EventKey:

    from_event = create_dispatch(
        on_init=lambda _: EventKey.init(),
        on_ready=lambda _: EventKey.ready(),
        on_offset=lambda _: EventKey.offset(),
        on_transaction_start=lambda _: EventKey.transaction_start(),
        on_transaction_end=lambda _: EventKey.transaction_end(),
        on_contract_created=lambda event: EventKey.contract_created(False, event.cid.template_id),
        on_contract_exercised=lambda event: EventKey.contract_exercised(
            False, event.cid.template_id, event.choice),
        on_contract_archived=lambda event: EventKey.contract_archived(False, event.cid.template_id))

    @staticmethod
    def init() -> Collection[str]:
        """
        Return the names of events that get raised in response to an :class:`InitEvent`. This is
        currently only ``'init'``.
        """
        return ('init',)

    @staticmethod
    def ready() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ReadyEvent`. This is
        currently only ``'ready'``.
        """
        return ('ready',)

    @staticmethod
    def offset() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`OffsetEvent`. This is
        currently only ``'offset'``.
        """
        return ('offset',)

    @staticmethod
    def transaction_start() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`TransactionStartEvent`.
        This is currently only ``'transaction-start'``.
        """
        return ('transaction-start',)

    @staticmethod
    def transaction_end() -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`TransactionEndEvent`.
        This is currently only ``'transaction-end'``.
        """
        return ('transaction-end',)

    @staticmethod
    def contract_created(primary_only: bool, template: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractCreateEvent`
        of the specified template type.
        """
        return EventKey._contract(primary_only, 'create', template)

    @staticmethod
    def contract_exercised(primary_only: bool, template: Any, choice: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractExercisedEvent`
        of the specified choice.
        """
        return [f'{key}/{choice}'
                for key in EventKey._contract(primary_only, 'exercised', template)]

    @staticmethod
    def contract_archived(primary_only: bool, template: Any) -> Collection[str]:
        """
        Return the names of events that get raised in response to a :class:`ContractCreateEvent`
        of the specified template type.
        """
        return EventKey._contract(primary_only, 'archive', template)

    @staticmethod
    def _contract(primary_only: bool, prefix: str, template: Any) -> Collection[str]:
        m, t = validate_template(template)
        return tuple(f'{prefix}/{g}' for g in template_reverse_globs(primary_only, m, t))


def sortable_offset_height(value: str) -> int:
    if value:
        components = value.split('-', 3)
        if len(components) == 1:
            return int(value)
        elif len(components) >= 1:
            return int(components[1])
    return 0



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains the logic for fuzzy lookups (specifically, template references) are performed
in dazl.
"""

from typing import Any, Iterator, Tuple


def validate_template(template: Any) -> Tuple[str, str]:
    """
    Return a module and type name component from something that can be interpreted as a template.

    :param template:
        Any object that can be interpreted as an identifier for a template.
    :return:
        A tuple of package ID and module name. The special value ``'*'`` is used if either the
        package ID, module name, or both should be wildcarded.
    :raise ValueError:
        If the object could not be interpreted as a thing referring to a template.
    """
    from .types import TemplateMeta, RecordType, TypeReference, UnresolvedTypeReference
    from .types_dynamic import ProxyMeta

    if template == '*' or template is None:
        return '*', '*'

    if isinstance(template, TemplateMeta):
        template = str(template)
    elif isinstance(type(template), ProxyMeta):
        template = str(template)
    elif isinstance(template, UnresolvedTypeReference):
        template = template.name

    if isinstance(template, str):
        if '@' in template:
            t, m = template.split('@')
            return m, t
        elif '/' in template:
            m, t = template.split('/')
            return m, t
        else:
            return '*', template
    elif isinstance(template, RecordType):
        template = template.name

    if isinstance(template, TypeReference):
        return template.module.package_id, template.full_name
    else:
        raise ValueError(f"Don't know how to convert {template!r} into a template")


def template_reverse_globs(primary_only: bool, module: str, type_name: str) -> Iterator[str]:
    """
    Return an iterator over strings that glob to a specified module's type.
    """
    module = module or '*'
    type_name = type_name or '*'

    if module != '*':
        if type_name != '*':
            yield f'{module}/{type_name}'
            if primary_only:
                return
        if not primary_only or type_name == '*':
            yield f'{module}/*'
    if type_name != '*':
        yield f'*/{type_name}'
        if primary_only:
            return
    yield '*/*'



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/active_contracts_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import event_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2
from . import transaction_filter_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__filter__pb2
from . import trace_context_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/active_contracts_service.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B ActiveContractsServiceOuterClass'),
  serialized_pb=_b('\n=com/digitalasset/ledger/api/v1/active_contracts_service.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a*com/digitalasset/ledger/api/v1/event.proto\x1a\x37\x63om/digitalasset/ledger/api/v1/transaction_filter.proto\x1a\x32\x63om/digitalasset/ledger/api/v1/trace_context.proto\"\xc8\x01\n\x19GetActiveContractsRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x41\n\x06\x66ilter\x18\x02 \x01(\x0b\x32\x31.com.digitalasset.ledger.api.v1.TransactionFilter\x12\x0f\n\x07verbose\x18\x03 \x01(\x08\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"\xcf\x01\n\x1aGetActiveContractsResponse\x12\x0e\n\x06offset\x18\x01 \x01(\t\x12\x13\n\x0bworkflow_id\x18\x02 \x01(\t\x12\x46\n\x10\x61\x63tive_contracts\x18\x03 \x03(\x0b\x32,.com.digitalasset.ledger.api.v1.CreatedEvent\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext2\xa8\x01\n\x16\x41\x63tiveContractsService\x12\x8d\x01\n\x12GetActiveContracts\x12\x39.com.digitalasset.ledger.api.v1.GetActiveContractsRequest\x1a:.com.digitalasset.ledger.api.v1.GetActiveContractsResponse0\x01\x42\x42\n\x1e\x63om.digitalasset.ledger.api.v1B ActiveContractsServiceOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2.DESCRIPTOR,com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__filter__pb2.DESCRIPTOR,com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,])




_GETACTIVECONTRACTSREQUEST = _descriptor.Descriptor(
  name='GetActiveContractsRequest',
  full_name='com.digitalasset.ledger.api.v1.GetActiveContractsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filter', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsRequest.filter', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='verbose', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsRequest.verbose', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsRequest.trace_context', index=3,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=251,
  serialized_end=451,
)


_GETACTIVECONTRACTSRESPONSE = _descriptor.Descriptor(
  name='GetActiveContractsResponse',
  full_name='com.digitalasset.ledger.api.v1.GetActiveContractsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsResponse.offset', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='workflow_id', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsResponse.workflow_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='active_contracts', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsResponse.active_contracts', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetActiveContractsResponse.trace_context', index=3,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=454,
  serialized_end=661,
)

_GETACTIVECONTRACTSREQUEST.fields_by_name['filter'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__filter__pb2._TRANSACTIONFILTER
_GETACTIVECONTRACTSREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETACTIVECONTRACTSRESPONSE.fields_by_name['active_contracts'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_event__pb2._CREATEDEVENT
_GETACTIVECONTRACTSRESPONSE.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
DESCRIPTOR.message_types_by_name['GetActiveContractsRequest'] = _GETACTIVECONTRACTSREQUEST
DESCRIPTOR.message_types_by_name['GetActiveContractsResponse'] = _GETACTIVECONTRACTSRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetActiveContractsRequest = _reflection.GeneratedProtocolMessageType('GetActiveContractsRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETACTIVECONTRACTSREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.active_contracts_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetActiveContractsRequest)
  ))
_sym_db.RegisterMessage(GetActiveContractsRequest)

GetActiveContractsResponse = _reflection.GeneratedProtocolMessageType('GetActiveContractsResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETACTIVECONTRACTSRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.active_contracts_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetActiveContractsResponse)
  ))
_sym_db.RegisterMessage(GetActiveContractsResponse)


DESCRIPTOR._options = None

_ACTIVECONTRACTSSERVICE = _descriptor.ServiceDescriptor(
  name='ActiveContractsService',
  full_name='com.digitalasset.ledger.api.v1.ActiveContractsService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=664,
  serialized_end=832,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetActiveContracts',
    full_name='com.digitalasset.ledger.api.v1.ActiveContractsService.GetActiveContracts',
    index=0,
    containing_service=None,
    input_type=_GETACTIVECONTRACTSREQUEST,
    output_type=_GETACTIVECONTRACTSRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_ACTIVECONTRACTSSERVICE)

DESCRIPTOR.services_by_name['ActiveContractsService'] = _ACTIVECONTRACTSSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module prints the metadata obtained from a remote server.
"""

from argparse import ArgumentParser
from typing import Collection

from ._base import CliCommand
from .. import LOG, Network
from ..client.config import configure_parser, get_config, merge_configurations, NetworkConfig
from ..model.types_store import PackageStore
from ..util.dar import DamlcPackageError
from ..util.dar_repo import LocalDarRepository
from ..pretty import get_pretty_printer, PrettyOptions


class PrintMetadataCommand(CliCommand):
    name = 'metadata'

    def parser(self) -> ArgumentParser:
        arg_parser = ArgumentParser()

        configure_parser(arg_parser)
        arg_parser.add_argument('--config', help='path to a YAML config file')
        arg_parser.add_argument('--file', help='path to a DAML or DAR file', action='append')
        arg_parser.add_argument('--show-hidden', help='show hidden types', action='store_true')
        arg_parser.add_argument('--format', help='one of \"daml\" or \"python\"', default='daml')
        return arg_parser

    def execute(self, args) -> int:
        LOG.debug('Executing a metadata fetch...')

        options = PrettyOptions(
            column_width=80,
            show_hidden_types=args.show_hidden,
            format=args.format)

        if args.file:
            return self.execute_static_metadata(args.file, options)

        config = get_config(args)

        return self.execute_runtime_metadata(config, options)

    @staticmethod
    def execute_static_metadata(files: Collection[str], options: PrettyOptions) -> int:
        repo = LocalDarRepository()
        try:
            repo.add_source(*files)
            _process_metadata(repo.store, options)
            return 0

        except DamlcPackageError as ex:
            return ex.exit_code

    @staticmethod
    def execute_runtime_metadata(config: 'NetworkConfig', options: PrettyOptions) -> int:
        network = Network()
        network.set_config(**config)
        global_info = network.simple_global()
        _process_metadata(global_info.metadata().store, options)
        return 0


def _process_metadata(store: PackageStore, options: PrettyOptions):
    import sys
    if sys.stdout.isatty():
        try:
            import pygments
            from pygments.formatters.terminal256 import Terminal256Formatter
            formatter = Terminal256Formatter()
        except ImportError:
            pygments = None
            formatter = None
    else:
        pygments = None
        formatter = None

    pretty_printer = get_pretty_printer(options.format, options, store)
    code = pretty_printer.render_store()
    lexer = pretty_printer.lexer()

    if pygments is not None and lexer is not None and formatter is not None:
        print(pygments.highlight(code, lexer, formatter))
    else:
        print(code)




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains utilities for deterministically outputting contract information.
"""

from collections import defaultdict


def collate_entries(metadata, parties, entries):
    """
    Return entries as a dict of :class:`TemplateMetadata` to list of :class:`LedgerCaptureEntry`.
    """
    sort = EntrySorter(parties)

    entries_by_template = defaultdict(list)
    for entry in entries:
        entries_by_template[entry.template_id].append(entry)

    grouped_entries = []
    for template_id, entries in entries_by_template.items():
        template = metadata.templates.find(template_id)
        entries.sort(key=sort.key)
        grouped_entries.append((template, entries))

    grouped_entries.sort(key=lambda t: t[0].template_name)
    return dict(grouped_entries)


class EntrySorter:
    """
    Implementation of a sort key method for :class:`LedgerCaptureEntry`.
    """
    def __init__(self, parties):
        self.parties = parties

    def key(self, entry):
        """
        Return an expression that can be used as an absolute sorting value as a proxy for this
        entry.
        """
        party_vis = [1 if entry.parties.get(party) is not None else 0 for party in self.parties]
        return sum(party_vis), ''.join(map(str, reversed(party_vis)))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Implementation of a codegen proxy type that can be used as a stand-in for codegen types when real
types are unavailable at the "right time".
"""

from dataclasses import dataclass
from typing import Any, Optional


@dataclass(frozen=True)
class NamedRecord:
    name: str
    arguments: dict


class ProxyMeta(type):
    """
    Metaclass for generated record types.
    """
    def __new__(mcs, name, bases, namespace, proxy_name: str):
        result = type.__new__(mcs, name, bases, dict(namespace))
        result._proxy_name = proxy_name
        return result

    def __str__(self):
        return self._proxy_name

    def __repr__(self):
        return self._proxy_name


def generated_type_proxy_root(package_id: 'Optional[str]' = None, root_module_name: 'str' = None) \
        -> Any:
    from typing import Dict
    from threading import RLock

    if not root_module_name:
        raise ValueError('root_module_name must be non-empty and specified')

    lock = RLock()
    types = dict()  # type: Dict[str, Any]

    def get_or_create_proxy(name: str):
        """
        Create a proxy type that looks like a record with a specified name. This name ultimately
        gets resolved later.
        """
        class GeneratedTypeProxy(metaclass=ProxyMeta, proxy_name=name):

            def __getattr__(self, item):
                """
                Create a new type reference with another component suffixed to the path of this
                component.
                """
                if not item:
                    raise AttributeError(item)

                return get_or_create_proxy(f'{name}.{item}')

            def __call__(self, *args, **kwargs):
                """
                Create a record-like structure.
                """
                if len(args):
                    # TODO: Handle the case where arguments are passed positionally. This is "hard"
                    #  because we don't necessarily have the metadata available to resolve
                    #  positional arguments to named arguments.
                    raise ValueError('positional arguments are not supported')
                return NamedRecord(name, kwargs)

            def __str__(self):
                """
                Return the name of the template.
                """
                return str(type(self))

        with lock:
            proxy = types.get(name)
            if proxy is None:
                proxy = types[name] = GeneratedTypeProxy()
            return proxy

    return get_or_create_proxy(root_module_name)





# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from ..model.network import OAuthSettings


async def oauth_flow(settings: OAuthSettings) -> OAuthSettings:
    """
    Implementation of an OAuth flow that ensures that a token is provided.

    :param settings:
    :return:
    """
    if not settings.token:
        raise ValueError('the token must be directly supplied at this time')
        # if settings.redirect_uri:
        #     client = WebApplicationClient(settings.client_id)
        #     client.prepare_request_uri()
        #     client.prepare_request_body()
        #     client.parse_request_uri_response()
        #     settings.
    return settings



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/commands.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import value_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/commands.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B\022CommandsOuterClass'),
  serialized_pb=_b('\n-com/digitalasset/ledger/api/v1/commands.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a*com/digitalasset/ledger/api/v1/value.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\x9c\x02\n\x08\x43ommands\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x13\n\x0bworkflow_id\x18\x02 \x01(\t\x12\x16\n\x0e\x61pplication_id\x18\x03 \x01(\t\x12\x12\n\ncommand_id\x18\x04 \x01(\t\x12\r\n\x05party\x18\x05 \x01(\t\x12\x39\n\x15ledger_effective_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x37\n\x13maximum_record_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x39\n\x08\x63ommands\x18\x08 \x03(\x0b\x32\'.com.digitalasset.ledger.api.v1.Command\"\xf1\x01\n\x07\x43ommand\x12?\n\x06\x63reate\x18\x01 \x01(\x0b\x32-.com.digitalasset.ledger.api.v1.CreateCommandH\x00\x12\x43\n\x08\x65xercise\x18\x02 \x01(\x0b\x32/.com.digitalasset.ledger.api.v1.ExerciseCommandH\x00\x12U\n\x11\x63reateAndExercise\x18\x03 \x01(\x0b\x32\x38.com.digitalasset.ledger.api.v1.CreateAndExerciseCommandH\x00\x42\t\n\x07\x63ommand\"\x92\x01\n\rCreateCommand\x12?\n\x0btemplate_id\x18\x01 \x01(\x0b\x32*.com.digitalasset.ledger.api.v1.Identifier\x12@\n\x10\x63reate_arguments\x18\x02 \x01(\x0b\x32&.com.digitalasset.ledger.api.v1.Record\"\xb7\x01\n\x0f\x45xerciseCommand\x12?\n\x0btemplate_id\x18\x01 \x01(\x0b\x32*.com.digitalasset.ledger.api.v1.Identifier\x12\x13\n\x0b\x63ontract_id\x18\x02 \x01(\t\x12\x0e\n\x06\x63hoice\x18\x03 \x01(\t\x12>\n\x0f\x63hoice_argument\x18\x04 \x01(\x0b\x32%.com.digitalasset.ledger.api.v1.Value\"\xed\x01\n\x18\x43reateAndExerciseCommand\x12?\n\x0btemplate_id\x18\x01 \x01(\x0b\x32*.com.digitalasset.ledger.api.v1.Identifier\x12@\n\x10\x63reate_arguments\x18\x02 \x01(\x0b\x32&.com.digitalasset.ledger.api.v1.Record\x12\x0e\n\x06\x63hoice\x18\x03 \x01(\t\x12>\n\x0f\x63hoice_argument\x18\x04 \x01(\x0b\x32%.com.digitalasset.ledger.api.v1.ValueB4\n\x1e\x63om.digitalasset.ledger.api.v1B\x12\x43ommandsOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])




_COMMANDS = _descriptor.Descriptor(
  name='Commands',
  full_name='com.digitalasset.ledger.api.v1.Commands',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.Commands.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='workflow_id', full_name='com.digitalasset.ledger.api.v1.Commands.workflow_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='application_id', full_name='com.digitalasset.ledger.api.v1.Commands.application_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='command_id', full_name='com.digitalasset.ledger.api.v1.Commands.command_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='party', full_name='com.digitalasset.ledger.api.v1.Commands.party', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ledger_effective_time', full_name='com.digitalasset.ledger.api.v1.Commands.ledger_effective_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='maximum_record_time', full_name='com.digitalasset.ledger.api.v1.Commands.maximum_record_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commands', full_name='com.digitalasset.ledger.api.v1.Commands.commands', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=159,
  serialized_end=443,
)


_COMMAND = _descriptor.Descriptor(
  name='Command',
  full_name='com.digitalasset.ledger.api.v1.Command',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='create', full_name='com.digitalasset.ledger.api.v1.Command.create', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exercise', full_name='com.digitalasset.ledger.api.v1.Command.exercise', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='createAndExercise', full_name='com.digitalasset.ledger.api.v1.Command.createAndExercise', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='command', full_name='com.digitalasset.ledger.api.v1.Command.command',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=446,
  serialized_end=687,
)


_CREATECOMMAND = _descriptor.Descriptor(
  name='CreateCommand',
  full_name='com.digitalasset.ledger.api.v1.CreateCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='template_id', full_name='com.digitalasset.ledger.api.v1.CreateCommand.template_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='create_arguments', full_name='com.digitalasset.ledger.api.v1.CreateCommand.create_arguments', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=690,
  serialized_end=836,
)


_EXERCISECOMMAND = _descriptor.Descriptor(
  name='ExerciseCommand',
  full_name='com.digitalasset.ledger.api.v1.ExerciseCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='template_id', full_name='com.digitalasset.ledger.api.v1.ExerciseCommand.template_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='contract_id', full_name='com.digitalasset.ledger.api.v1.ExerciseCommand.contract_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='choice', full_name='com.digitalasset.ledger.api.v1.ExerciseCommand.choice', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='choice_argument', full_name='com.digitalasset.ledger.api.v1.ExerciseCommand.choice_argument', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=839,
  serialized_end=1022,
)


_CREATEANDEXERCISECOMMAND = _descriptor.Descriptor(
  name='CreateAndExerciseCommand',
  full_name='com.digitalasset.ledger.api.v1.CreateAndExerciseCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='template_id', full_name='com.digitalasset.ledger.api.v1.CreateAndExerciseCommand.template_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='create_arguments', full_name='com.digitalasset.ledger.api.v1.CreateAndExerciseCommand.create_arguments', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='choice', full_name='com.digitalasset.ledger.api.v1.CreateAndExerciseCommand.choice', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='choice_argument', full_name='com.digitalasset.ledger.api.v1.CreateAndExerciseCommand.choice_argument', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1025,
  serialized_end=1262,
)

_COMMANDS.fields_by_name['ledger_effective_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COMMANDS.fields_by_name['maximum_record_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COMMANDS.fields_by_name['commands'].message_type = _COMMAND
_COMMAND.fields_by_name['create'].message_type = _CREATECOMMAND
_COMMAND.fields_by_name['exercise'].message_type = _EXERCISECOMMAND
_COMMAND.fields_by_name['createAndExercise'].message_type = _CREATEANDEXERCISECOMMAND
_COMMAND.oneofs_by_name['command'].fields.append(
  _COMMAND.fields_by_name['create'])
_COMMAND.fields_by_name['create'].containing_oneof = _COMMAND.oneofs_by_name['command']
_COMMAND.oneofs_by_name['command'].fields.append(
  _COMMAND.fields_by_name['exercise'])
_COMMAND.fields_by_name['exercise'].containing_oneof = _COMMAND.oneofs_by_name['command']
_COMMAND.oneofs_by_name['command'].fields.append(
  _COMMAND.fields_by_name['createAndExercise'])
_COMMAND.fields_by_name['createAndExercise'].containing_oneof = _COMMAND.oneofs_by_name['command']
_CREATECOMMAND.fields_by_name['template_id'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
_CREATECOMMAND.fields_by_name['create_arguments'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2._RECORD
_EXERCISECOMMAND.fields_by_name['template_id'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
_EXERCISECOMMAND.fields_by_name['choice_argument'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2._VALUE
_CREATEANDEXERCISECOMMAND.fields_by_name['template_id'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2._IDENTIFIER
_CREATEANDEXERCISECOMMAND.fields_by_name['create_arguments'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2._RECORD
_CREATEANDEXERCISECOMMAND.fields_by_name['choice_argument'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_value__pb2._VALUE
DESCRIPTOR.message_types_by_name['Commands'] = _COMMANDS
DESCRIPTOR.message_types_by_name['Command'] = _COMMAND
DESCRIPTOR.message_types_by_name['CreateCommand'] = _CREATECOMMAND
DESCRIPTOR.message_types_by_name['ExerciseCommand'] = _EXERCISECOMMAND
DESCRIPTOR.message_types_by_name['CreateAndExerciseCommand'] = _CREATEANDEXERCISECOMMAND
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Commands = _reflection.GeneratedProtocolMessageType('Commands', (_message.Message,), dict(
  DESCRIPTOR = _COMMANDS,
  __module__ = 'com.digitalasset.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.Commands)
  ))
_sym_db.RegisterMessage(Commands)

Command = _reflection.GeneratedProtocolMessageType('Command', (_message.Message,), dict(
  DESCRIPTOR = _COMMAND,
  __module__ = 'com.digitalasset.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.Command)
  ))
_sym_db.RegisterMessage(Command)

CreateCommand = _reflection.GeneratedProtocolMessageType('CreateCommand', (_message.Message,), dict(
  DESCRIPTOR = _CREATECOMMAND,
  __module__ = 'com.digitalasset.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.CreateCommand)
  ))
_sym_db.RegisterMessage(CreateCommand)

ExerciseCommand = _reflection.GeneratedProtocolMessageType('ExerciseCommand', (_message.Message,), dict(
  DESCRIPTOR = _EXERCISECOMMAND,
  __module__ = 'com.digitalasset.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.ExerciseCommand)
  ))
_sym_db.RegisterMessage(ExerciseCommand)

CreateAndExerciseCommand = _reflection.GeneratedProtocolMessageType('CreateAndExerciseCommand', (_message.Message,), dict(
  DESCRIPTOR = _CREATEANDEXERCISECOMMAND,
  __module__ = 'com.digitalasset.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.CreateAndExerciseCommand)
  ))
_sym_db.RegisterMessage(CreateAndExerciseCommand)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/package_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import trace_context_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/package_service.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B\030PackageServiceOuterClass'),
  serialized_pb=_b('\n4com/digitalasset/ledger/api/v1/package_service.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a\x32\x63om/digitalasset/ledger/api/v1/trace_context.proto\"n\n\x13ListPackagesRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"+\n\x14ListPackagesResponse\x12\x13\n\x0bpackage_ids\x18\x01 \x03(\t\"\x80\x01\n\x11GetPackageRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x12\n\npackage_id\x18\x02 \x01(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"\x80\x01\n\x12GetPackageResponse\x12\x43\n\rhash_function\x18\x01 \x01(\x0e\x32,.com.digitalasset.ledger.api.v1.HashFunction\x12\x17\n\x0f\x61rchive_payload\x18\x02 \x01(\x0c\x12\x0c\n\x04hash\x18\x03 \x01(\t\"\x86\x01\n\x17GetPackageStatusRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x12\n\npackage_id\x18\x02 \x01(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"a\n\x18GetPackageStatusResponse\x12\x45\n\x0epackage_status\x18\x01 \x01(\x0e\x32-.com.digitalasset.ledger.api.v1.PackageStatus*,\n\rPackageStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0e\n\nREGISTERED\x10\x01*\x1a\n\x0cHashFunction\x12\n\n\x06SHA256\x10\x00\x32\x88\x03\n\x0ePackageService\x12y\n\x0cListPackages\x12\x33.com.digitalasset.ledger.api.v1.ListPackagesRequest\x1a\x34.com.digitalasset.ledger.api.v1.ListPackagesResponse\x12s\n\nGetPackage\x12\x31.com.digitalasset.ledger.api.v1.GetPackageRequest\x1a\x32.com.digitalasset.ledger.api.v1.GetPackageResponse\x12\x85\x01\n\x10GetPackageStatus\x12\x37.com.digitalasset.ledger.api.v1.GetPackageStatusRequest\x1a\x38.com.digitalasset.ledger.api.v1.GetPackageStatusResponseB:\n\x1e\x63om.digitalasset.ledger.api.v1B\x18PackageServiceOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,])

_PACKAGESTATUS = _descriptor.EnumDescriptor(
  name='PackageStatus',
  full_name='com.digitalasset.ledger.api.v1.PackageStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REGISTERED', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=795,
  serialized_end=839,
)
_sym_db.RegisterEnumDescriptor(_PACKAGESTATUS)

PackageStatus = enum_type_wrapper.EnumTypeWrapper(_PACKAGESTATUS)
_HASHFUNCTION = _descriptor.EnumDescriptor(
  name='HashFunction',
  full_name='com.digitalasset.ledger.api.v1.HashFunction',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SHA256', index=0, number=0,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=841,
  serialized_end=867,
)
_sym_db.RegisterEnumDescriptor(_HASHFUNCTION)

HashFunction = enum_type_wrapper.EnumTypeWrapper(_HASHFUNCTION)
UNKNOWN = 0
REGISTERED = 1
SHA256 = 0



_LISTPACKAGESREQUEST = _descriptor.Descriptor(
  name='ListPackagesRequest',
  full_name='com.digitalasset.ledger.api.v1.ListPackagesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.ListPackagesRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.ListPackagesRequest.trace_context', index=1,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=140,
  serialized_end=250,
)


_LISTPACKAGESRESPONSE = _descriptor.Descriptor(
  name='ListPackagesResponse',
  full_name='com.digitalasset.ledger.api.v1.ListPackagesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='package_ids', full_name='com.digitalasset.ledger.api.v1.ListPackagesResponse.package_ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=252,
  serialized_end=295,
)


_GETPACKAGEREQUEST = _descriptor.Descriptor(
  name='GetPackageRequest',
  full_name='com.digitalasset.ledger.api.v1.GetPackageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetPackageRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='package_id', full_name='com.digitalasset.ledger.api.v1.GetPackageRequest.package_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetPackageRequest.trace_context', index=2,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=298,
  serialized_end=426,
)


_GETPACKAGERESPONSE = _descriptor.Descriptor(
  name='GetPackageResponse',
  full_name='com.digitalasset.ledger.api.v1.GetPackageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash_function', full_name='com.digitalasset.ledger.api.v1.GetPackageResponse.hash_function', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='archive_payload', full_name='com.digitalasset.ledger.api.v1.GetPackageResponse.archive_payload', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hash', full_name='com.digitalasset.ledger.api.v1.GetPackageResponse.hash', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=429,
  serialized_end=557,
)


_GETPACKAGESTATUSREQUEST = _descriptor.Descriptor(
  name='GetPackageStatusRequest',
  full_name='com.digitalasset.ledger.api.v1.GetPackageStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetPackageStatusRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='package_id', full_name='com.digitalasset.ledger.api.v1.GetPackageStatusRequest.package_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetPackageStatusRequest.trace_context', index=2,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=560,
  serialized_end=694,
)


_GETPACKAGESTATUSRESPONSE = _descriptor.Descriptor(
  name='GetPackageStatusResponse',
  full_name='com.digitalasset.ledger.api.v1.GetPackageStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='package_status', full_name='com.digitalasset.ledger.api.v1.GetPackageStatusResponse.package_status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=696,
  serialized_end=793,
)

_LISTPACKAGESREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETPACKAGEREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETPACKAGERESPONSE.fields_by_name['hash_function'].enum_type = _HASHFUNCTION
_GETPACKAGESTATUSREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETPACKAGESTATUSRESPONSE.fields_by_name['package_status'].enum_type = _PACKAGESTATUS
DESCRIPTOR.message_types_by_name['ListPackagesRequest'] = _LISTPACKAGESREQUEST
DESCRIPTOR.message_types_by_name['ListPackagesResponse'] = _LISTPACKAGESRESPONSE
DESCRIPTOR.message_types_by_name['GetPackageRequest'] = _GETPACKAGEREQUEST
DESCRIPTOR.message_types_by_name['GetPackageResponse'] = _GETPACKAGERESPONSE
DESCRIPTOR.message_types_by_name['GetPackageStatusRequest'] = _GETPACKAGESTATUSREQUEST
DESCRIPTOR.message_types_by_name['GetPackageStatusResponse'] = _GETPACKAGESTATUSRESPONSE
DESCRIPTOR.enum_types_by_name['PackageStatus'] = _PACKAGESTATUS
DESCRIPTOR.enum_types_by_name['HashFunction'] = _HASHFUNCTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListPackagesRequest = _reflection.GeneratedProtocolMessageType('ListPackagesRequest', (_message.Message,), dict(
  DESCRIPTOR = _LISTPACKAGESREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.ListPackagesRequest)
  ))
_sym_db.RegisterMessage(ListPackagesRequest)

ListPackagesResponse = _reflection.GeneratedProtocolMessageType('ListPackagesResponse', (_message.Message,), dict(
  DESCRIPTOR = _LISTPACKAGESRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.ListPackagesResponse)
  ))
_sym_db.RegisterMessage(ListPackagesResponse)

GetPackageRequest = _reflection.GeneratedProtocolMessageType('GetPackageRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETPACKAGEREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetPackageRequest)
  ))
_sym_db.RegisterMessage(GetPackageRequest)

GetPackageResponse = _reflection.GeneratedProtocolMessageType('GetPackageResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETPACKAGERESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetPackageResponse)
  ))
_sym_db.RegisterMessage(GetPackageResponse)

GetPackageStatusRequest = _reflection.GeneratedProtocolMessageType('GetPackageStatusRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETPACKAGESTATUSREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetPackageStatusRequest)
  ))
_sym_db.RegisterMessage(GetPackageStatusRequest)

GetPackageStatusResponse = _reflection.GeneratedProtocolMessageType('GetPackageStatusResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETPACKAGESTATUSRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.package_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetPackageStatusResponse)
  ))
_sym_db.RegisterMessage(GetPackageStatusResponse)


DESCRIPTOR._options = None

_PACKAGESERVICE = _descriptor.ServiceDescriptor(
  name='PackageService',
  full_name='com.digitalasset.ledger.api.v1.PackageService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=870,
  serialized_end=1262,
  methods=[
  _descriptor.MethodDescriptor(
    name='ListPackages',
    full_name='com.digitalasset.ledger.api.v1.PackageService.ListPackages',
    index=0,
    containing_service=None,
    input_type=_LISTPACKAGESREQUEST,
    output_type=_LISTPACKAGESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetPackage',
    full_name='com.digitalasset.ledger.api.v1.PackageService.GetPackage',
    index=1,
    containing_service=None,
    input_type=_GETPACKAGEREQUEST,
    output_type=_GETPACKAGERESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetPackageStatus',
    full_name='com.digitalasset.ledger.api.v1.PackageService.GetPackageStatus',
    index=2,
    containing_service=None,
    input_type=_GETPACKAGESTATUSREQUEST,
    output_type=_GETPACKAGESTATUSRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_PACKAGESERVICE)

DESCRIPTOR.services_by_name['PackageService'] = _PACKAGESERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Generic, TypeVar
from .daml_lf_1 import Expr, Type, ValName, BuiltinFunction, PrimCon, PrimLit, Case, Block, Update, \
    Scenario, Package, Module, FieldWithType, FieldWithExpr, VarWithType, Binding

E = TypeVar('E', contravariant=True)
RP = TypeVar('RP', covariant=True)
RM = TypeVar('RM', covariant=True)
RE = TypeVar('RE', covariant=True)
RT = TypeVar('RT', covariant=True)
_tuple = tuple


class PackageVisitor(Generic[RP]):
    def visit_package(self, package: 'Package') -> 'RP':
        raise NotImplementedError


class ModuleVisitor(Generic[RM]):
    def visit_module(self, module: 'Module') -> 'RM':
        raise NotImplementedError


class ExprVisitor(Generic[RE]):

    def visit_expr(self, expr: 'Expr') -> 'RE':
        return expr.Sum_match(
            self.visit_expr_var,
            self.visit_expr_val,
            self.visit_expr_builtin,
            self.visit_expr_prim_con,
            self.visit_expr_prim_lit,
            self.visit_expr_rec_con,
            self.visit_expr_rec_proj,
            self.visit_expr_variant_con,
            self.visit_expr_tuple_con,
            self.visit_expr_tuple_proj,
            self.visit_expr_app,
            self.visit_expr_ty_app,
            self.visit_expr_abs,
            self.visit_expr_ty_abs,
            self.visit_expr_case,
            self.visit_expr_let,
            self.visit_expr_nil,
            self.visit_expr_cons,
            self.visit_expr_update,
            self.visit_expr_scenario,
            self.visit_expr_rec_upd,
            self.visit_expr_tuple_upd,
            self.visit_expr_none,
            self.visit_expr_some)

    def visit_expr_var(self, var: str) -> 'RE':
        raise NotImplementedError

    def visit_expr_val(self, val: 'ValName') -> 'RE':
        raise NotImplementedError

    def visit_expr_builtin(self, builtin: 'BuiltinFunction') -> 'RE':
        raise NotImplementedError

    def visit_expr_prim_con(self, prim_con: 'PrimCon') -> 'RE':
        raise NotImplementedError

    def visit_expr_prim_lit(self, prim_lit: 'PrimLit') -> 'RE':
        raise NotImplementedError

    def visit_expr_rec_con(self, rec_con: 'Expr.RecCon') -> 'RE':
        raise NotImplementedError

    def visit_expr_rec_proj(self, rec_proj: 'Expr.RecProj') -> 'RE':
        raise NotImplementedError

    def visit_expr_variant_con(self, variant_con: 'Expr.VariantCon') -> 'RE':
        raise NotImplementedError

    def visit_expr_tuple_con(self, tuple_con: 'Expr.TupleCon') -> 'RE':
        raise NotImplementedError

    def visit_expr_tuple_proj(self, tuple_proj: 'Expr.TupleProj') -> 'RE':
        raise NotImplementedError

    def visit_expr_app(self, app: 'Expr.App') -> 'RE':
        raise NotImplementedError

    def visit_expr_ty_app(self, ty_app: 'Expr.TyApp') -> 'RE':
        raise NotImplementedError

    def visit_expr_abs(self, abs_: 'Expr.Abs') -> 'RE':
        raise NotImplementedError

    def visit_expr_ty_abs(self, ty_abs: 'Expr.TyAbs') -> 'RE':
        raise NotImplementedError

    def visit_expr_case(self, case: 'Case') -> 'RE':
        raise NotImplementedError

    def visit_expr_let(self, let: 'Block') -> 'RE':
        raise NotImplementedError

    def visit_expr_nil(self, nil: 'Expr.Nil') -> 'RE':
        raise NotImplementedError

    def visit_expr_cons(self, cons: 'Expr.Cons') -> 'RE':
        raise NotImplementedError

    def visit_expr_update(self, update: 'Update') -> 'RE':
        raise NotImplementedError

    def visit_expr_scenario(self, scenario: 'Scenario') -> 'RE':
        raise NotImplementedError

    def visit_expr_rec_upd(self, rec_upd: 'Expr.RecUpd') -> 'RE':
        raise NotImplementedError

    def visit_expr_tuple_upd(self, tuple_upd: 'Expr.TupleUpd') -> 'RE':
        raise NotImplementedError

    def visit_expr_none(self, none: 'Expr.None_') -> 'RE':
        raise NotImplementedError

    def visit_expr_some(self, some: 'Expr.Some') -> 'RE':
        raise NotImplementedError


class TypeVisitor(Generic[RT]):

    def visit_type(self, type: 'Type') -> 'RT':
        return type.Sum_match(
            self.visit_type_var,
            self.visit_type_con,
            self.visit_type_prim,
            self.visit_type_forall,
            self.visit_type_tuple)

    def visit_type_var(self, var: 'Type.Var') -> 'RT':
        raise NotImplementedError

    def visit_type_con(self, con: 'Type.Con') -> 'RT':
        raise NotImplementedError

    def visit_type_prim(self, prim: 'Type.Prim') -> 'RT':
        raise NotImplementedError

    def visit_type_forall(self, forall: 'Type.Forall') -> 'RT':
        raise NotImplementedError

    def visit_type_tuple(self, tuple: 'Type.Tuple') -> 'RT':
        raise NotImplementedError


class IdentityTypeVisitor(TypeVisitor[Type]):

    def resolve_type(self, var: 'str') -> 'Optional[Type]':
        """
        Return a :class:`Type` for the specified variable, or ``None`` if there is no mapping for
        that type variable.
        """
        raise NotImplementedError

    def visit_type_var(self, var: 'Type.Var') -> 'Type':
        resolved_type = self.resolve_type(var.var)
        if resolved_type is not None:
            base_type = self.visit_type(resolved_type)
        else:
            base_type = Type(var=var)

        if var.args:
            new_args = tuple(self.visit_type(arg) for arg in var.args)
            return Type(var=Type.Var(var=var.var, args=new_args))
        else:
            return base_type

    def visit_type_con(self, con: 'Type.Con') -> 'Type':
        new_args = tuple(self.visit_type(arg) for arg in con.args)
        return Type(con=Type.Con(tycon=con.tycon, args=new_args))

    def visit_type_prim(self, prim: 'Type.Prim') -> 'Type':
        new_args = tuple(self.visit_type(arg) for arg in prim.args)
        return Type(prim=Type.Prim(prim=prim.prim, args=new_args))

    def visit_type_forall(self, forall: 'Type.Forall') -> 'Type':
        new_body = self.visit_type(forall.body)
        return Type(forall=Type.Forall(forall.vars, new_body))

    def visit_type_tuple(self, tuple: 'Type.Tuple') -> 'Type':
        new_fields = _tuple([FieldWithType(field=fwt.field, type=self.visit_type(fwt.type))
                             for fwt in tuple.fields])
        return Type(tuple=Type.Tuple(fields=new_fields))


class IdentityVisitor(ExprVisitor[Expr], IdentityTypeVisitor):
    """
    Implementation of :class:`ExprVisitor` that returns the same object for every case.
    """

    def visit_expr(self, expr: 'Expr') -> 'Expr':
        return super(IdentityVisitor, self).visit_expr(expr)

    def visit_expr_var(self, var: str) -> 'Expr':
        return Expr(var=var)

    def visit_expr_val(self, val: 'ValName') -> 'Expr':
        return Expr(val=val)

    def visit_expr_builtin(self, builtin: 'BuiltinFunction') -> 'Expr':
        return Expr(builtin=builtin)

    def visit_expr_prim_con(self, prim_con: 'PrimCon') -> 'Expr':
        return Expr(prim_con=prim_con)

    def visit_expr_prim_lit(self, prim_lit: 'PrimLit') -> 'Expr':
        return Expr(prim_lit=prim_lit)

    def visit_expr_rec_con(self, rec_con: 'Expr.RecCon') -> 'Expr':
        new_fields = tuple([FieldWithExpr(fwt.field, self.visit_expr(fwt.expr))
                            for fwt in rec_con.fields])
        new_tycon = self.visit_type_con(rec_con.tycon).con
        return Expr(rec_con=Expr.RecCon(tycon=new_tycon, fields=new_fields))

    def visit_expr_rec_proj(self, rec_proj: 'Expr.RecProj') -> 'Expr':
        new_tycon = self.visit_type_con(rec_proj.tycon).con
        new_record = self.visit_expr(rec_proj.record)
        return Expr(rec_proj=Expr.RecProj(tycon=new_tycon, field=rec_proj.field, record=new_record))

    def visit_expr_variant_con(self, variant_con: 'Expr.VariantCon') -> 'Expr':
        new_type_con = self.visit_type_con(variant_con.tycon).con
        new_variant_arg = self.visit_expr(variant_con.variant_arg)
        return Expr(variant_con=Expr.VariantCon(
            tycon=new_type_con, variant_con=variant_con.variant_con, variant_arg=new_variant_arg))

    def visit_expr_tuple_con(self, tuple_con: 'Expr.TupleCon') -> 'Expr':
        new_fields = tuple([FieldWithExpr(fwt.field, self.visit_expr(fwt.expr))
                            for fwt in tuple_con.fields])
        return Expr(tuple_con=Expr.TupleCon(fields=new_fields))

    def visit_expr_tuple_proj(self, tuple_proj: 'Expr.TupleProj') -> 'Expr':
        new_tuple = self.visit_expr(tuple_proj.tuple)
        return Expr(tuple_proj=Expr.TupleProj(field=tuple_proj.field, tuple=new_tuple))

    def visit_expr_app(self, app: 'Expr.App') -> 'Expr':
        new_fun = self.visit_expr(app.fun)
        new_args = tuple([self.visit_expr(arg) for arg in app.args])
        return Expr(app=Expr.App(new_fun, new_args))

    def visit_expr_ty_app(self, ty_app: 'Expr.TyApp') -> 'Expr':
        new_expr = self.visit_expr(ty_app.expr)
        new_types = tuple([self.visit_type(t) for t in ty_app.types])
        return Expr(ty_app=Expr.TyApp(new_expr, new_types))

    def visit_expr_abs(self, abs_: 'Expr.Abs') -> 'Expr':
        new_body = self.visit_expr(abs_.body)
        return Expr(abs=Expr.Abs(param=abs_.param, body=new_body))

    def visit_expr_ty_abs(self, ty_abs: 'Expr.TyAbs') -> 'Expr':
        new_body = self.visit_expr(ty_abs.body)
        return Expr(ty_abs=Expr.TyAbs(param=ty_abs.param, body=new_body))

    def visit_expr_case(self, case: 'Case') -> 'Expr':
        new_scrut = self.visit_expr(case.scrut)
        # TODO: Real implementation
        new_alts = [alt for alt in case.alts]
        return Expr(case=Case(scrut=new_scrut, alts=new_alts))

    def visit_expr_let(self, let: 'Block') -> 'Expr':
        new_bindings = [Binding(VarWithType(binding.binder.var, self.visit_type(binding.binder.type)),
                                self.visit_expr(binding.bound)) for binding in let.bindings]
        new_body = self.visit_expr(let.body)
        return Expr(let=Block(bindings=new_bindings, body=new_body))

    def visit_expr_nil(self, nil: 'Expr.Nil') -> 'Expr':
        return Expr(nil=Expr.Nil(self.visit_type(nil.type)))

    def visit_expr_cons(self, cons: 'Expr.Cons') -> 'Expr':
        new_type = self.visit_type(cons.type)
        new_front = tuple(self.visit_expr(front) for front in cons.front)
        new_tail = self.visit_expr(cons.tail)
        return Expr(cons=Expr.Cons(type=new_type, front=new_front, tail=new_tail))

    def visit_expr_update(self, update: 'Update') -> 'Expr':
        return Expr(update=update)

    def visit_expr_scenario(self, scenario: 'Scenario') -> 'Expr':
        return Expr(scenario=scenario)

    def visit_expr_rec_upd(self, rec_upd: 'Expr.RecUpd') -> 'Expr':
        return Expr(rec_upd=rec_upd)

    def visit_expr_tuple_upd(self, tuple_upd: 'Expr.TupleUpd') -> 'Expr':
        return Expr(tuple_upd=tuple_upd)

    def visit_expr_none(self, none: 'Expr.None_') -> 'Expr':
        new_type = self.visit_type(none.type)
        return Expr(none=Expr.None_(type=new_type))

    def visit_expr_some(self, some: 'Expr.Some') -> 'Expr':
        new_type = self.visit_type(some.type)
        new_body = self.visit_expr(some.body)
        return Expr(some=Expr.Some(type=new_type, body=new_body))




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from io import StringIO
from typing import Optional, Sequence, Union

from ._render_base import PrettyPrintBase
from .util import maybe_parentheses
from ..damlast.daml_lf_1 import DefDataType, DefTemplate, Expr, Module, PrimCon, PrimType, \
    Type as NewType
from ..model.types import Type as OldType, ScalarType, ContractIdType, ListType, OptionalType, \
    MapType, RecordType, TypeApp, TypeVariable, TypeReference, UpdateType, VariantType, \
    type_dispatch_table, SCALAR_TYPE_UNIT, SCALAR_TYPE_BOOL, SCALAR_TYPE_CHAR, \
    SCALAR_TYPE_INTEGER, SCALAR_TYPE_DECIMAL, SCALAR_TYPE_TEXT, SCALAR_TYPE_PARTY, \
    SCALAR_TYPE_RELTIME, SCALAR_TYPE_DATE, SCALAR_TYPE_TIME, ForAllType


_OldTypePrim = Union[ScalarType, ContractIdType, ListType, OptionalType, MapType, UpdateType,
                     NewType.Prim]


class DamlPrettyPrinter(PrettyPrintBase):

    syntax = 'daml'

    def lexer(self):
        try:
            from .pygments_daml_lexer import DAMLLexer
        except ImportError:
            return None

        return DAMLLexer()

    def visit_module_ref_start(self, module_ref: 'ModuleRef'):
        module_name = '.'.join(module_ref.module_name)
        return f'module {module_name} where\n'

    def visit_def_data_type(self, def_data_type: 'DefDataType') -> str:
        return self.visit_def_template(None, def_data_type)

    def visit_def_template(
            self, template: 'Optional[DefTemplate]', def_data_type: 'DefDataType') -> str:
        """
        Render a template (or a simple data type).

        :param template:
            The :class:`DefTemplate` to render. If not supplied, only the :class:`DefDataType` is
            rendered.
        :param def_data_type:
            The :class:`DefDataType` to render.
        :return:
            String rendering of the template and/or data type.
        """
        local_name = '.'.join(def_data_type.name.segments)

        from ..damlast.expand import ExpandVisitor, SimplifyVisitor
        ex = ExpandVisitor(self.store)
        sp = SimplifyVisitor(self.store)

        def render(expr: Expr) -> str:
            #return self.visit_expr(expr)
            e1 = ex.visit_expr(expr)
            e2 = sp.visit_expr(e1)
            return self.visit_expr(e2)

        with StringIO() as buf:
            if def_data_type.record is not None:
                type_args = ''.join(f' {p.var}' for p in def_data_type.params)
                if template is not None:
                    buf.write('template ' + local_name)
                else:
                    buf.write(f'data {local_name}{type_args} = {local_name}')
                buf.write('\n  with\n')
                for field in def_data_type.record.fields:
                    buf.write(f'    {field.field}: {self.visit_type(field.type)}\n')
                if template is not None:
                    buf.write('  where\n')
                    buf.write(f'    signatory {render(template.signatories)}\n')
                    buf.write(f'    observer {render(template.observers)}\n')
                    buf.write(f'    ensure {render(template.precond)}\n')
                    buf.write(f'    agreement {render(template.agreement)}\n')
                    # buf.write('    controller')
            elif def_data_type.variant is not None:
                buf.write('data ' + local_name + ' = ')

            if template is None:
                if def_data_type.serializable:
                    buf.write('\n  deriving (Eq, Show)')

            return buf.getvalue()

    # region visit_expr_* methods

    def visit_expr_prim_con_inline_unit(self):
        return '()'

    def visit_expr_rec_con_decl(self, rec_con: 'Expr.RecCon', context: 'CodeContext'):
        return self._visit_expr_decl(
            context, self.visit_expr_rec_con_inline(rec_con, line_endings=True))

    def visit_expr_rec_con_inline(self, rec_con: 'Expr.RecCon', line_endings: bool = False):
        ctor_expr = f'{self.visit_type_con(rec_con.tycon)} with'
        field_exprs = [f'{fwt.field}={self.visit_expr(fwt.expr)}' for fwt in rec_con.fields]
        if line_endings:
            return f'{ctor_expr} {"; ".join(field_exprs)}'
        else:
            return ctor_expr + '\n  '.join(field_exprs)

    def visit_expr_rec_proj(self, rec_proj: 'Expr.RecProj'):
        rec_text = maybe_parentheses(self.visit_expr(rec_proj.record))
        return f'{rec_text}.{rec_proj.field}'

    def visit_expr_variant_con(self, variant_con: 'Expr.VariantCon') -> str:
        arg_text = self.visit_expr(variant_con.variant_arg)
        return f'{variant_con.variant_con} {maybe_parentheses(arg_text)}'

    def visit_expr_tuple_proj(self, tuple_proj: 'Expr.TupleProj'):
        tuple_text = maybe_parentheses(self.visit_expr(tuple_proj.tuple))
        return f'{tuple_text}.{tuple_proj.field}'

    def visit_expr_app_inline(self, app: 'Expr.App'):
        components = [app.fun, *app.args]
        return ' '.join(maybe_parentheses(self.visit_expr(e)) for e in components)

    def visit_expr_abs_decl(self, abs_: 'Expr.Abs', context: 'CodeContext'):
        local_name = '.'.join(context.decl_name)
        field_def = f'{local_name} :: {self.visit_type(context.decl_type)}'
        all_params = [t.var for t in context.type_abs] if context.type_abs is not None else []
        all_params.extend(p.var for p in abs_.param)

        param_text = ''.join(' ' + p for p in all_params)
        body_text = self.visit_expr(abs_.body)
        field_decl = f'{local_name}{param_text} = {body_text}'
        return f'{field_def}\n{field_decl}'

    def visit_expr_abs_inline(self, abs_: 'Expr.Abs'):
        param_text = ' '.join(p.var for p in abs_.param)
        body_expr = self.visit_expr(abs_.body)
        return f'(\\{param_text} -> {body_expr})'

    def visit_expr_case(self, case: 'Case'):
        scrut_text = maybe_parentheses(self.visit_expr(case.scrut))
        patterns = ['  ' + self.visit_expr_casealt(alt) for alt in case.alts]
        lines = [f'case {scrut_text} of']
        lines.extend(patterns)
        return '\n'.join(lines)

    def visit_expr_let(self, let: 'Block') -> str:
        binding_lines = [f'{binding.binder.var} = {self.visit_expr(binding.bound)}'
                         for binding in let.bindings]
        with StringIO() as buf:
            buf.write('let ')
            delim = ''
            for line in binding_lines:
                buf.write(delim)
                buf.write(line)
                buf.write('\n')
                delim = '    '
            buf.write('in ')
            expr_text = self.visit_expr(let.body)
            buf.write(expr_text)
            return buf.getvalue()

    def visit_expr_nil_inline(self, nil: 'Expr.Nil') -> str:
        return 'nil'

    def visit_expr_cons(self, cons: 'Expr.Cons') -> str:
        front_texts = [self.visit_expr(front) for front in cons.front]
        if cons.tail.nil:
            return '[' + ', '.join(maybe_parentheses(t, ',') for t in front_texts) + ']'

        tail_text = self.visit_expr(cons.tail)
        if len(front_texts) > 1:
            return '[' + ', '.join(maybe_parentheses(t, ',') for t in front_texts) + '] ++ ' + \
                   maybe_parentheses(tail_text)
        else:
            return f'{front_texts[0]} `Cons` {tail_text}'

    def visit_expr_update(self, update: 'Update'):
        return 'do ... return'

    def visit_expr_scenario(self, scenario: 'Scenario'):
        return 'test ... return'

    def visit_expr_none(self, none: 'Expr.None_') -> str:
        return f'None'

    def visit_expr_some(self, some: 'Expr.Some') -> str:
        expr_text = self.visit_expr(some.body)
        return f'Some {maybe_parentheses(expr_text)}'

    # endregion

    # region visit_type_* methods

    def visit_type(self, type: 'Union[str, NewType, OldType]', parenthesize: bool = False) -> str:
        if isinstance(type, str):
            type_str = type
        elif isinstance(type, NewType):
            type_str = type.Sum_match(
                var=self.visit_type_var,
                con=self.visit_type_con,
                prim=self.visit_type_prim,
                forall=self.visit_type_forall,
                tuple=self.visit_type_tuple)
        elif isinstance(type, UpdateType):
            type_str = self.visit_type_prim(type)
        elif isinstance(type, ForAllType):
            type_str = self.visit_type_forall(type)
        elif isinstance(type, OldType):
            type_str = type_dispatch_table(
                on_type_var=self.visit_type_var,
                on_type_ref=self.visit_type_con,
                on_scalar=self.visit_type_prim,
                on_contract_id=self.visit_type_prim,
                on_list=self.visit_type_prim,
                on_optional=self.visit_type_prim,
                on_map=self.visit_type_prim,
                on_record=self.visit_type_con,
                on_variant=self.visit_type_con,
                on_type_app=self._visit_type_app,
                on_unsupported=repr)(type)
        else:
            raise TypeError(f'A DAML Type is required here (got {type!r} instead')

        return maybe_parentheses(type_str) if parenthesize else type_str

    def visit_type_var(self, var: 'Union[str, TypeVariable, NewType.Var]'):
        if isinstance(var, NewType.Var):
            return self._visit_type_app((var.var, *var.args))
        elif isinstance(var, TypeVariable):
            return var.name
        elif isinstance(var, str):
            return var
        else:
            raise TypeError(f'A DAML Type variable is required here (got {var!r} instead')

    def visit_type_con(self, con: 'Union[TypeReference, RecordType, VariantType, NewType.Con]') -> str:
        if isinstance(con, TypeReference):
            return con.full_name_unambiguous
        elif isinstance(con, (RecordType, VariantType)):
            if con.name is None:
                raise ValueError('A named Record or Variant type is required here')
            return con.name.full_name_unambiguous
        elif isinstance(con, NewType.Con):
            return self._visit_type_app((con.tycon.full_name_unambiguous, *con.args))
        else:
            raise TypeError(f'A DAML Type constructor is required here (got {con!r} instead')

    def visit_type_prim(self, prim: '_OldTypePrim') -> str:
        prim_type = prim.prim if isinstance(prim, NewType.Prim) else None

        if PrimType.UNIT == prim_type or SCALAR_TYPE_UNIT == prim:
            return 'Unit'
        elif PrimType.BOOL == prim_type or SCALAR_TYPE_BOOL == prim:
            return 'Bool'
        elif PrimType.INT64 == prim_type or SCALAR_TYPE_INTEGER == prim:
            return 'Int'
        elif PrimType.DECIMAL == prim_type or SCALAR_TYPE_DECIMAL == prim:
            return 'Decimal'
        elif PrimType.CHAR == prim_type or SCALAR_TYPE_CHAR == prim:
            return 'Char'
        elif PrimType.TEXT == prim_type or SCALAR_TYPE_TEXT == prim:
            return 'Text'
        elif PrimType.TIMESTAMP == prim_type or SCALAR_TYPE_TIME == prim:
            return 'Time'
        elif PrimType.RELTIME == prim_type or SCALAR_TYPE_RELTIME == prim:
            return 'RelTime'
        elif PrimType.PARTY == prim_type or SCALAR_TYPE_PARTY == prim:
            return 'Party'

        elif PrimType.LIST == prim_type:
            if len(prim.args) == 0:
                return 'List'
            else:
                return f'[{self.visit_type(prim.args[0])}]'
        elif isinstance(prim, ListType):
            return f'[{self.visit_type(prim.type_parameter)}]'

        elif PrimType.UPDATE == prim_type:
            return self._visit_type_app(('Update', *prim.args))
        elif isinstance(prim, UpdateType):
            return self._visit_type_app(('Update', prim.type_parameter))

        elif PrimType.SCENARIO == prim_type:
            return self._visit_type_app(('Scenario', *prim.args))

        elif PrimType.DATE == prim_type or SCALAR_TYPE_DATE == prim:
            return 'Date'

        elif PrimType.CONTRACT_ID == prim_type:
            return self._visit_type_app(('ContractId', *prim.args))
        elif isinstance(prim, ContractIdType):
            return self._visit_type_app(('ContractId', prim.type_parameter))

        elif PrimType.OPTIONAL == prim_type:
            return self._visit_type_app(('Optional', *prim.args))
        elif isinstance(prim, OptionalType):
            return self._visit_type_app(('Optional', prim.type_parameter))

        elif PrimType.ARROW == prim_type:
            arrow_operator = ' -> '
            return arrow_operator.join(
                maybe_parentheses(self.visit_type(a), arrow_operator) for a in prim.args)

        elif PrimType.MAP == prim_type:
            return self._visit_type_app(('Map', *prim.args))
        elif PrimType.MAP_GENERIC == prim_type:
            return self._visit_type_app(('Map', *prim.args))
        elif isinstance(prim, MapType):
            return self._visit_type_app(('Map', prim.key_type, prim.value_type))

        else:
            raise TypeError(f'A DAML Type primitive is required here (got {prim!r} instead')

    def visit_type_forall(self, forall: 'Union[ForAllType, NewType.Forall]') -> str:
        if isinstance(forall, NewType.Forall):
            return f'forall {" ".join(self.visit_type_var(v.var) for v in forall.vars)}. {self.visit_type(forall.body)}'
        elif isinstance(forall, ForAllType):
            return f'forall {" ".join(self.visit_type_var(v.var) for v in forall.type_vars)}. ' \
                f'{self.visit_type(forall.body_type)}'
        else:
            raise TypeError(f'A DAML forall Type is required here (got {forall!r} instead')

    # noinspection PyShadowingBuiltins
    def visit_type_tuple(self, tuple: 'NewType.Tuple') -> str:
        return '(' + ','.join(self.visit_type(t.type) for t in tuple.fields) + ')'

    def _visit_type_app(self, types: 'Union[TypeApp, Sequence[Union[str, NewType, OldType]]]'):
        if isinstance(types, TypeApp):
            types = (types.body, *types.arguments)

        return ' '.join(self.visit_type(c, True) for c in types)

    # endregion


DEFAULT_PRINTER = DamlPrettyPrinter()



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from argparse import ArgumentParser

from .. import LOG
from ._base import CliCommand
from ..plugins import LedgerCapturePlugin
from ..client.manager import LedgerClientManager
from ..client.config import configure_parser, get_config


class ListAllCommand(CliCommand):
    name = 'ls'

    def parser(self) -> ArgumentParser:
        arg_parser = ArgumentParser('dazl ls')

        configure_parser(arg_parser, config_file_support=True)
        arg_parser.add_argument('--format', '--fmt', '-F', type=str, default=LedgerCapturePlugin.DEFAULT_FORMATTER_NAME)
        arg_parser.add_argument('--template-filter', '-T', type=str)
        arg_parser.add_argument('--all', '-A', action='store_true')
        return arg_parser

    def execute(self, args) -> int:
        LOG.debug('Executing an ls...')

        final_config = get_config(args, config_file_support=True)

        with LedgerClientManager(final_config) as mgr:
            kwargs = dict()
            if args.template_filter:
                kwargs['template_filter'] = [template.strip() for template in args.template_filter.split(',')]
            if args.all:
                kwargs['include_archived'] = True

            inspector = LedgerCapturePlugin.stdout(**kwargs)
            mgr.register(inspector)
            mgr.run_until_complete()

            inspector.dump_all(fmt=args.format)
            return 0




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import active_contracts_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2


class ActiveContractsServiceStub(object):
  """Allows clients to initialize themselves according to a fairly recent state of the ledger without reading through all transactions that were committed since the ledger's creation.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetActiveContracts = channel.unary_stream(
        '/com.digitalasset.ledger.api.v1.ActiveContractsService/GetActiveContracts',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsResponse.FromString,
        )


class ActiveContractsServiceServicer(object):
  """Allows clients to initialize themselves according to a fairly recent state of the ledger without reading through all transactions that were committed since the ledger's creation.
  """

  def GetActiveContracts(self, request, context):
    """Returns a stream of the latest snapshot of active contracts. Getting an empty stream means that the active contracts set is empty and the client should listen to transactions using ``LEDGER_BEGIN``.
    Clients SHOULD NOT assume that the set of active contracts they receive reflects the state at the ledger end.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ActiveContractsServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetActiveContracts': grpc.unary_stream_rpc_method_handler(
          servicer.GetActiveContracts,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_active__contracts__service__pb2.GetActiveContractsResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.ActiveContractsService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from urllib.request import Request, urlopen

from ... import LOG


def admin_upload_package(url: str, contents: bytes) -> None:
    request = Request(url, method='PUT')
    with urlopen(request, contents) as response:
        status = response.getcode()
        if 200 <= status <= 299:
            return None
        else:
            LOG.error('Received %s when trying to upload a DALF to %s', status, url)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from argparse import ArgumentParser

from ._base import CliCommand
from ..damlsdk import ensure_sdk_component, SDKComponentNotFoundError, SDKComponentUnknownError


class FetchComponentCommand(CliCommand):
    name = 'fetch'
    hidden = True

    def parser(self):
        arg_parser = ArgumentParser('dazl fetch')
        arg_parser.add_argument('component', metavar='NAME:VERSION', help='name:version of the component to fetch')
        arg_parser.add_argument('--force',
                                help='download and unpack artifacts regardless of whether they exist',
                                action='store_true')
        return arg_parser

    def execute(self, args) -> int:
        component = args.component

        try:
            cmps = component.split(':')
            if len(cmps) == 2:
                result = ensure_sdk_component(cmps[0], cmps[1], force=args.force)
            elif len(cmps) == 1:
                result = ensure_sdk_component(cmps[0], None, force=args.force)
            else:
                print('Invalid component identifier: {component!r}. Must be of the form <component>:<version>.')
                return 4
            if result.downloaded:
                print('Successfully downloaded:')
            else:
                print('Skipped downloading because it already exists locally:')
            print(f'    {result.url}')
            print(f'to')
            print(f'    {result.path}')
            return 0

        except SDKComponentNotFoundError:
            print(f'Could not download {component!r}')
            return 2
        except SDKComponentUnknownError:
            print(f'Don\'t know what {component!r} is.')
            print('Only "sdk", "damlc:<version>" and "sandbox:<version>" are supported.')
            return 3

        return 0



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Pretty print format that orders templates by ID, and attempts to render a more compact
representation.
"""

import re

from collections import defaultdict
from typing import Iterable, Optional

from .model_capture import LedgerCapture, LedgerCaptureEntry
from dazl.util.termcap import termsize
from ...model.types import Type, RecordType

BOX_B3 = '|'
BOX_C4 = '-'
BOX_DA = '+'


# a list of template parameter names that are long and make it tough to understand
# what is going on when written out as a single line
_COMMON_PREFIXES = ['invitedAs']


class _TemplateKey:
    """
    Type that identifies templates for display purposes.
    """
    @staticmethod
    def unknown(identifier: str) -> '_TemplateKey':
        return _TemplateKey(identifier, None)

    @staticmethod
    def for_record_type(record_type: RecordType) -> '_TemplateKey':
        return _TemplateKey(
            '.'.join(record_type.name.module.module_name + record_type.name.name),
            record_type)

    def __init__(self, template_name: str, data_type: Optional[Type]):
        self.template_name = template_name
        self.data_type = data_type


def format_error(error: str) -> Iterable[str]:
    return [error]


def format_entries(capture: LedgerCapture,
                   parties: Iterable[str],
                   entries: Optional[Iterable[LedgerCaptureEntry]]=None,
                   color: bool=True,
                   columns: Optional[int]=None) -> Iterable[str]:
    if columns is None:
        _, columns = termsize()
        if columns is None:
            columns = 120

    sort = ByPartySort(parties)

    entries_by_template = defaultdict(list)
    if entries is None:
        entries = capture

    for entry in entries:
        entries_by_template[entry.template_id].append(entry)

    grouped_entries = []
    for template_id, entries in entries_by_template.items():
        # try to resolve a template for the template ID
        template_type = None
        if capture.store is not None:
            candidates = capture.store.resolve_template_type(template_id)
            if len(candidates) == 1:
                template_type = next(iter(candidates.values()))

        # if we can successfully resolve metadata and render instances of this template nicely,
        # then use that rendering; otherwise resort to a rougher display
        key = _TemplateKey.for_record_type(template_type) if template_type is not None \
            else _TemplateKey.unknown(template_id)
        grouped_entries.append((key, entries))

    grouped_entries.sort(key=lambda t: t[0].template_name)

    entry_count = sum(len(entries) for entries in entries_by_template.values())

    yield "{} total contracts over {} templates".format(entry_count, len(entries_by_template))
    yield from party_header(parties)
    for template, entries in grouped_entries:
        # for each column for all entries in the data set, determine the most
        # compact representation here
        yield ''

        renderer = _TemplateEntryRenderer(template, parties)
        renderer.measure(entries)

        yield from renderer.render_header(color=color, columns=columns)
        yield from renderer.render_entries(sorted(entries, key=sort.key))


def party_header(parties):
    parties = tuple(parties)
    party_count = len(parties)

    for index, party in enumerate(parties):
        ascii_art = (BOX_B3 * index) + BOX_DA + (BOX_C4 * (party_count - index - 1))
        yield f"{ascii_art} party '{party}'"
    yield (('|' * party_count))


class ByPartySort:
    def __init__(self, parties):
        self.parties = parties

    def key(self, entry):
        party_vis = [1 if entry.parties.get(party) is not None else 0 for party in self.parties]
        return (sum(party_vis), ''.join(map(str, reversed(party_vis))))


def split_header_name(name, max_length=None):
    """
    Splits a name in a header so that it is as short as possible while still being readable.
    """
    candidate = [name]
    for prefix in _COMMON_PREFIXES:
        if name.startswith(prefix):
            candidate = [prefix, name[len(prefix):]]
            break

    if max_length is not None and any(len(line) > max_length for line in candidate):
        # this is still a little long; try slicing on all uppercase letters
        split_half_index = 1 if len(candidate) >= 1 else 0

        substitution = candidate[0:split_half_index]
        if split_half_index < len(candidate):
            next_upper = re.match('[A-Z]', candidate[split_half_index])
            if next_upper is not None:
                idx = next_upper.start(0)
                if idx > 0:
                    substitution.append(candidate[split_half_index][0:idx])
                for line in candidate[split_half_index:]:
                    substitution.extend(re.findall('[A-Z][^A-Z]*', line))
        return substitution

    return candidate


class _TemplateEntryRenderer:
    """
    A template-specific renderer.
    """

    def __init__(self, template: _TemplateKey, parties):
        self.max_widths = {}
        if template.data_type is not None:
            self.headers = [_Header(name, param, parties)
                            for name, param in template.data_type.as_args_list()]
            self.template_name = template.template_name
        else:
            self.headers = [_Header(".cdata", None, parties)]
            self.template_name = template.template_name

        self.headers.insert(0, _Header(".cid", None, parties))
        self.headers.insert(0, _Header(".party", None, parties, colsize=len(parties)))
        self.entry_count = 0

    def measure(self, entries):
        for entry in entries:
            self.entry_count += 1
            for header in self.headers:
                header.measure_cell(entry)

    def render_header(self, color, columns=None):
        """
        Return an iterable over header rows.
        """
        if self.entry_count == 1:
            template_header_row = '{!s} (1 contract)'.format(self.template_name)
        else:
            template_header_row = '{!s} ({!s} contracts)'.format(self.template_name, self.entry_count)

        if columns is not None and len(template_header_row) < (columns - 1):
            yield template_header_row + ' ' + ''.rjust(columns - len(template_header_row) - 1, '-')
        else:
            yield template_header_row

        header_row_height = max(len(header.header_lines) for header in self.headers)
        for i in range(header_row_height):
            yield ' '.join(header.header_cell(i, header_row_height) for header in self.headers)

    def render_entries(self, entries):
        """
        Return an iterable that renders entries in order of their visibility to parties.
        """
        for entry in entries:
            yield ' '.join(header.render_cell(entry) for header in self.headers)


class _Header:
    def __init__(self, name, param, parties, colsize=None):
        self.name = name

        if name == '.cid':
            self.header_lines = ['#cid']
            self._value_from_entry = lambda entry: entry.contract_id
        elif name == '.party':
            self.header_lines = ['']
            self._value_from_entry = lambda entry: render_parties(parties, entry)
        elif name == '.cdata':
            self.header_lines = ['#cdata (metadata missing)']
            self._value_from_entry = lambda entry: entry.contract_args
        else:
            self.header_lines = split_header_name(name, 8 if hasattr(param, 'name') and param.name == 'Bool' else None)
            self._value_from_entry = lambda entry: entry.contract_args[name] if entry.contract_args is not None else None

        self.colsize = colsize if colsize is not None else max(len(header) for header in self.header_lines)

    def header_cell(self, index, count):
        offset = count - len(self.header_lines)
        actual_index = index - offset
        return (self.header_lines[actual_index] if actual_index >= 0 else '').ljust(self.colsize, ' ')

    def measure_cell(self, data):
        data_size = len(str(self._value_from_entry(data)))
        self.colsize = max(self.colsize, data_size)

    def render_cell(self, entry):
        return str(self._value_from_entry(entry)).ljust(self.colsize, ' ')

    def __repr__(self):
        return '<_Header(name={!r}, colsize={!r})>'.format(self.name, self.colsize)


def render_parties(all_parties, entry):
    return ''.join(_render_party_bool(entry.parties.get(party)) for party in all_parties)


def _render_party_bool(value):
    if value is not None:
        return 'C' if value else 'A'
    return ' '



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains the Python API for interacting with the Ledger API.
"""


import pkg_resources as _pkg_resources
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .util.logging import LoggerWithVerbose


def _initialize_logger() -> 'LoggerWithVerbose':
    import logging
    from .util.logging import _LoggerWithVerbose, log_verbose

    try:
        # noinspection PyPackageRequirements
        import google

        # This method is undocumented, but is required to read large size of model files when using
        # the C++ implementation.
        # noinspection PyPackageRequirements,PyUnresolvedReferences,PyProtectedMember
        from google.protobuf.pyext import _message
        _message.SetAllowOversizeProtos(True)

    except ImportError:
        pass

    original_logger = logging.getLoggerClass()

    class CombinedLogger(original_logger, _LoggerWithVerbose):
        pass

    logging.setLoggerClass(CombinedLogger)
    dazl_logger = logging.getLogger('dazl')
    if not hasattr(dazl_logger, 'verbose'):
        from functools import partial
        dazl_logger.verbose = partial(log_verbose, dazl_logger)
    logging.setLoggerClass(original_logger)
    return dazl_logger


LOG = _initialize_logger()  # type: LoggerWithVerbose


from .damlsdk.sandbox import sandbox  # noqa
from .client import create_client, simple_client, Network  # noqa
from .model.core import DazlError  # noqa
from .model.types import module, TemplateMeta, ChoiceMeta  # noqa
from .model.writing import CreateCommand, ExerciseCommand, create, exercise  # noqa
from .util.logging import setup_default_logger  # noqa
from .util.prim_types import frozendict  # noqa
from .plugins.capture.plugin_capture import write_acs  # noqa


__version__ = _pkg_resources.require('dazl')[0].version
